use proc_macro::TokenStream;
use syn::{parse::Parse, parse_macro_input, ItemFn, LitInt, Token};
use quote::quote;

struct NativeFuncArgs {
	arity: LitInt,
	variadic: Option<Token![..]>,
}

impl Parse for NativeFuncArgs {
	fn parse(input: syn::parse::ParseStream) -> syn::Result<Self> {
		let arity = input.parse()?;
		let variadic = input.parse()?;
		Ok(Self { arity, variadic })
	}
}

#[proc_macro_attribute]
pub fn native_func(input: TokenStream, annotated_item: TokenStream) -> TokenStream {
	let Ok(itemfn) = syn::parse::<ItemFn>(annotated_item.clone()) else {
		return annotated_item
	};
	let args: NativeFuncArgs = parse_macro_input!(input as NativeFuncArgs);

	let visibility = itemfn.vis;
	let block = itemfn.block;
	let name = itemfn.sig.ident;
	let inputs = itemfn.sig.inputs;
	let output = itemfn.sig.output;
	let arity = args.arity;
	let variadic = args.variadic.is_some();

	assert!(itemfn.sig.constness.is_none(), "item must not be const");
	assert!(itemfn.sig.asyncness.is_none(), "item must not be async");
	assert!(itemfn.sig.unsafety.is_none(), "item must not be unsafe");
	assert!(itemfn.sig.abi.is_none(), "item must not contain an ABI specifier");
	assert!(itemfn.sig.variadic.is_none(), "item must not be variadic");

	let expanded = quote! {
		#visibility fn #name() -> ::talc_lang::value::function::NativeFunc {
			::talc_lang::value::function::NativeFunc {
				attrs: ::talc_lang::value::function::FuncAttrs{
					arity: #arity,
					variadic: #variadic,
				},
				func: Box::new(|#inputs| #output #block)
			}
		}
	};

	TokenStream::from(expanded)
}
